%Control of floor heating process

%Specialization project 2012
%Siri Hofstad Trapnes

%Simple model

close all 
clear all
clc

%% Simulation parameters

%Data

l_wall = 2.4; %[m]
l_floor = 5; %[m]
A_floor = 25; %[m^2]

x_wood = 0.015; %[m]  %Assumption: Wood = oak
x_rock_wool = 0.25; %[m]
x_concrete = 0.1; %[m]

%Densitiy
d_concrete = 2400; %[kg/m^3]
d_wood = 825; %[kg/m^3]
d_rock_wool = 192; %[kg/m^3]
d_air = 1.166; %[kg/m^3] average between 20 and 40C 

%Heat capacity
Cp_concrete = 0.63*1000; %[J/kgK]  
Cpr = 1.005*1000; % [J/kgK] Cp_air at -50 - 40C
Cpf = Cp_concrete;

%Heat transfer coefficients
hair = 15*60*60; %[J/(h x m^2 x K)] (Still air)
k_wood = 0.208*60*60; %[J/(h x m x K)] Oak
k_rock_wool = 0.0391*60*60; %[J/(h x m x K)]  At 37.8C
k_concrete = 0.762*60*60; %[J/(h x m x K)]

% ------------------------------------------------------------------------

%Mass = density x volume
%Floor
m_concrete = d_concrete*(x_concrete*A_floor);
%m_wood = d_wood*(x_wood*A_floor);
mf = m_concrete; % [kg]
%Room
%m_wood1 = d_wood*(x_wood*l_wall*l_floor);
%m_rock_wool = d_rock_wool*(x_rock_wool*l_wall*l_floor);
mr = d_air*(l_wall*A_floor); % [kg]

% ------------------------------------------------------------------------

%Floor
Af = A_floor; % [m^2]
Ufr = 1/(x_wood/k_wood); %[J/(h x m^2 x K)]

UAfr = Ufr*Af;
mCpf = mf*Cpf;

%Room
Ar = 4*l_wall*l_floor; % [m^2] Varme  fra fire vegger
Uro = 1/((1/hair)+(x_wood/k_wood)+(x_rock_wool/k_rock_wool)+...
 (x_wood/k_wood)+(1/hair)); %0.18*60*60; %[J/(h x m^2 x K)] 

UAro = Uro*Ar*10;
mCpr = mr*Cpr;


%% Inputs and initial conditions

%Inputs
qhr = 4000*60*60; % No heat on: 0 %max: 4000*60*60; % [J/h]
qhf = 0; %No heat on: 0 %max: 575*60; % [J/h]
To = 273+0; % [K]

%Initial conditions
Tf = 283; % [K] 
Tr = 283; % [K]
E = 0;%Energy consumption  [J/h] (integratinng (qhf + qhr))
init = [Tf Tr E];

%Set point
Trset = 293; % [K]
Tmin = 283; % [K]
Tmax = 293; % [K]


%% Transfer functions of the system

%State space model
format long
A = [((-UAfr)/mCpf) (UAfr/mCpf) 0; (UAfr/mCpr) ((-UAro-UAfr)/mCpr) 0; 0 0 0];  
B = [1/mCpf 0 0 ; 0 1/mCpr (UAro/mCpr); 0 1 0];
C = [1 0 0; 0 1 0; 0 0 1];
D = [0 0 0 ; 0 0 0; 0 0 0];

lsys = ss(A,B,C,D);

%Transfer functions
G = tf(lsys);



%% Simulation without PI controlller

sim('simulink_simple')

%Outputs
Tf = simout(:,1)-273; % [C]
Tr = simout(:,2)-273; % [C]

%Simulation time
t = time; %[h] 1 day

%Plotting

figure(1)
Subplot(2,1,1)
plot(t,Tf,'b')
ylabel('Temperature [C]')
xlabel('Time [h]')
title('Tf') 

Subplot(2,1,2)
plot(t,Tr,'r')
ylabel('Temperature [C]')
xlabel('Time [h]')
title('Tr') 

%% Simulation with PI controller

sim('simulink_simple_PI')

%Outputs
Tf = simout(:,1)-273; % [C]
Tr = simout(:,2)-273; % [C]
Tr_set = Set_out-273; % [C]
qhr_out = Qhr;
To_out = To_out-273; % [C]
E = E_out; 

%Simulation time
t = time; %[h] 1 day

% Plotting

%Tr and qhr
figure(2)
Subplot(2,1,1)
plot(t,Tr,'r')
hold on
plot(t,Tr_set,'b')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('Tr','Tr set') 

Subplot(2,1,2)
plot(t,qhr_out)
ylabel('qhr [J/h]')
xlabel('Time [h]')
title('Room')

%Outdoor temperature
figure(3)
plot(t,To_out,'g')
ylabel('Outdoor temperature, To [C]')
xlabel('Time [h]')

%Energy consumption
figure(4)
plot(t,E_out,'r')
ylabel('Energy consumption [J/h]')
xlabel('Time [h]')

%% Optimal dynamical control problem

%% Input

k = t;  %One day, 24h
h = 1/50; 
%Min temperaures at different time intervals
Tmin_a = 283; % [K]
Tmin_b = 293; % [K]
Tmin_c = 283; % [K]
Tmin_d = 293; % [K]
Tmin_e = 283; % [K]

%Lower and upper bounds
Tr_upper = 400; % [K]
Tf_upper = 400; % [K]
qhf_upper = 0; % [J/h]
qhr_upper = 4000*60*60; % [J/h]
To_upper = To; % [K]
E_upper = Inf; % [J/h]
qhf_lower = 0; % [J/h]
qhr_lower = 0; % [J/h]
To_lower = To; % [K]
E_lower = 0; % [J/h]
Tf_lower = 277; % [K] 
Tr_lower = 277; % [K] 

% -----------------------------------------------------------------------

tstart = t(1);
tend = t(end);
n = (tend-tstart)/h;

% -----------------------------------------------------------------------

%Time intervals corresponding to the different minimum temperatures    
%     t(7.5) - tstart
n1 = n/5; %length(tstart:h:t(38)) %(t(38)-tstart)/h   
%     t(8.5) - t(7.5)
n2 = n/5; %length(t(52):h:t(38)) %(t(52)-t(38))/h  
%     t(16) - t(8.5)
n3 = n/5; %length(t(127):h:t(52)) %(t(127)-t(52))/h  
%     t(23) - t(16)
n4 = n/5; %length(t(153):h:t(127)) %(t(153)-t(127))/h 
%     tend - t(23)
n5 = n/5; %length(tend:h:t(153)) %(tend-t(153))/h  


%% Minimazation

%Discrete model
% x_k+1 = Ad*x_k + Bd*u_k
% y_k+1 = Cd*x_k  

Ad = h*lsys.a+eye(3);
Bd = h*lsys.b;  
Cd = lsys.c; 
Dd = D;

[nx, nu] = size(lsys.b);
[ny, ~]  = size(lsys.c);

% Want the system on the form a*z = b

%Matrix b
b = zeros(n,1);   
b(1) = 1;
xo = init';
b_1 = (Ad)*xo;
b = kron(b,b_1); %Dimension 2n x 1
%Matrix a
a_0 = ones(n);  
%a_1 = tril(a_0,-1);
%a_2 = tril(a_0,-2);
%a_3 = a_1-a_2;
a_3 = diag(ones(1,n-1),-1);
a_4 = kron(a_3,-Ad);
a1 = eye(nx*n)+a_4;  %(2*n)
a2 = kron(eye(n),-Bd); 
a = horzcat(a1,a2); 

%Quadprog solves the following problem
% min_x o.5*x'*H*x+f'*x
%subject to
%A x <= b  Inequality constrints
%Aeq x = beq  Equality consrtaints
%Lb, Ub  Lower and upper bounds

%Cost function
%Matrix H
H = zeros((nu+nx)*n,(nu+nx)*n);   
%Matrix f
f_0 = ones(1,n);    
f_1 = zeros(1,nx*n);      
f_2 = kron(f_0,[0 1 0]); %[Tf Tr E]
f = horzcat(f_1,f_2);
f = f';             %Sum qhr   


%Constraints

%Equality constraints
Aeq = a; 
beq = b;

%Inequality constraints
%Matrix A  (= c)
%                  Tf Tr E 
c_0 = kron(eye(n),[0  1  0]);    
c_1 = zeros(n,nu*n);    
c = (-1)*horzcat(c_0,c_1);   

%Matrix B  (= d)
d = (-1)*[ones(n1,1)*(Tmin_a); % day 1   
     ones(n2,1)*(Tmin_b);
     ones(n3,1)*(Tmin_c);
     ones(n4,1)*(Tmin_d); 
     ones(n5,1)*(Tmin_e)]; % [K]       

d1 = [ones(n1,1)*(Tmin_a)-273; 
     ones(n2,1)*(Tmin_b)-273;
     ones(n3,1)*(Tmin_c)-273;
     ones(n4,1)*(Tmin_d)-273; 
     ones(n5,1)*(Tmin_e)-273]; % [C]

 
%Upper and lower bounds
%Lower
LB = kron(ones(n,1),[Tf_lower Tr_lower E_lower]'); 
LB_2 = kron(ones(n,1),[qhf_lower qhr_lower To_lower]'); 
LB = [LB;LB_2];
%Upper
ub_1 = kron(ones(n,1),[Tf_upper Tr_upper E_upper]');
ub_2 = kron(ones(n,1),[qhf_upper qhr_upper To_upper]'); 
UB = vertcat(ub_1,ub_2);  
opt = optimset('MaxIter',200,'display','iter','algorithm','active-set')
%Z gives the optimal values of X (Tr, Tf, E) and U (qhr, qhf, To)
[Z,FVAL,EXITFLAG,OUTPUT,LAMBDA] = QUADPROG(H,f,c,d,Aeq,beq,LB,UB,[],opt);
H1=H;

%% Plotting

%Outputs
X = Z(1:nx*n);
% Take every third X. Get Tf in the first row, Tr in the second and E in
% the third
x = reshape(X,3,[]);
U = Z(nx*n+1:end);
% Take every third U. Get qfr in the first row, qhr in the second and To in
% the third
u = reshape(U,3,[]);

x_Tf = [x(1,:)-273]'; % [C]
x_Tr = [x(2,:)-273]'; % [C]
x_E = [x(3,:)]';
u_qhf = u(1,:)'; 
u_qhr = u(2,:)';
u_To = [u(3,:)-273]'; % [C]

t1 = [t(1)+h:h:t(end)]';   


% %Optimal Tr and qhr
figure(5)
Subplot(2,1,1)
plot(t1,x_Tr,'r')
hold on
plot(t1,d1,'b')
title('Optimal temperature')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('Tr','Tr constraint')
Subplot(2,1,2)
plot(t1,u_qhr,'b')
title('Optimal qhr')
ylabel('qhr [J/h]')
xlabel('Time [h]')
legend('qhr')  
 
%Optimal Tf, qhf and To
figure(6)
subplot(3,1,1)
plot(t1,x_Tf,'r')
title('Optimal temperature')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('Tf') 
Subplot(3,1,2)
plot(t1,u_qhf,'b')
title('Optimal qhf')
ylabel('qhf [J/h]')
xlabel('Time [h]')
legend ('qhf') 
subplot(3,1,3)
plot(t1,u_To,'g')
title('Optimal temperature')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('To') 

%Optimal energy consumption
figure(7)
plot(t1,x_E,'r')
title('Optimal energy consumption')
ylabel('Energy consumption [J/h]')
xlabel('Time [h]')

%% Energy consumption

figure(8)
plot(t,E_out,'r')
hold on
plot(t1,x_E,'b')
legend('E','E_{opt}') 
ylabel('Energy consumption [J/h]')
xlabel('Time [h]')

%% Creating a plot with opt Tr and opt qhr, 
% but also including the constraints from the optimalization. 

%Opt system
figure(9)
Subplot(2,1,1)
plot(t1,x_Tr,'r')
hold on
%plot(t,Tr_set,'b')
%hold on
plot(t1,d1,'g')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('Tr','Tr set', 'Tr constraint') 

Subplot(2,1,2)
plot(t1,u_qhr)
ylabel('qhr [J/h]')
xlabel('Time [h]')
title('Room')

%% Nullspace method

%Nominal values 
To_n = 273; % [K]
t_n = 3.66; % [h]
DeltaT_n = 10; % [K] Tr-To
qhr_n = 2.59*10^6; % [J/h]
di_n = sqrt(2); %Distance
t_sqrt_n = 1.0677; % sqrt(|tc-ts|) 
tsq_n = 1.9131; % sqrt(ts)
yn = [To_n t_n DeltaT_n qhr_n di_n t_sqrt_n  tsq_n]'; %

%Disturbance 1
To_d1 = 268; % [K]
t_d1 = 3.04; % [h]
DeltaT_d1 = 15; % [K]
qhr_d1 = 3.885*10^6; % [J/h]
di_d1 = 1.854;
t_sqrt_d1 = 1.334166;
tsq_d1 = 1.74356;
yd1_1 = [To_d1 t_d1 DeltaT_d1 qhr_d1 di_d1 t_sqrt_d1 tsq_d1 ]'; % 

%Disturbance 2
% To_d2 = 273; % [K]
% t_d2 = 4.06; % [h]
% DeltaT_d2 = 10; % [K]
% qhr_d2 = 2.072*10^6; % [J/h]
% di_d2 = 1.202; 
% t_sqrt_d2 = 0.8717798;
% tsq_d2 = 2.0149;
% yd2_1 = [To_d2 t_d2 DeltaT_d2 qhr_d2 di_d2 t_sqrt_d2 tsq_d2 ]'; %

%Disturbance 3
To_d3 = 277; % [K]
t_d3 = 4.06; % [h]
DeltaT_d3 = 6; % [K]
qhr_d3 = 1.544*10^6; % [J/h]
di_d3 = 1.733; 
t_sqrt_d3 = 0.8717798;
tsq_d3 = 2.0149;
yd3_1 = [To_d3 t_d3 DeltaT_d3 qhr_d3 di_d3 t_sqrt_d3 tsq_d3]';

%Values used when scaling di. Nominal values.
Ts_n = 283; % [K]
ts_n = 3.66; % [h]
Tc = 293; % [K]
tc = t1(240); % [h]
Tc_n = Tc; % [K]
tc_n = tc; % [h]

%Scaling the measurements: (y-yn)/ymax
%       To   t  DeltaT     qhr      di    t_sqrt    tsq    
ymax = [5  0.4   15    1.44*10^7  0.4396 0.266466 0.1018]'; %Delta ymax
%           0.4                     1.8
yd1 = (yd1_1-yn)./ymax; %Disturbance 1
%yd2 = (yd2_1-yn)./ymax; %Disturbance 2
yd3 = (yd3_1-yn)./ymax; %Disturbance 3

%Scaling the disturbances: (d-dn)/dmax
Tomax = 5; %Outdoor temperature (d1)
Tomax_2 =4; %Outdoor temperature (d3)
UAro_n = 2.5902*10^5; % [J/h]
UAro_d2 = 2.0721*10^5; % [J/h]
UAmax = 0.5*UAro_n;

dmax = [Tomax UAmax Tomax];

delta_d1 = (To_d1-To_n)/dmax(1); 
%delta_d2 = (UAro_d2-UAro_n)/dmax(2); 
delta_d3 = (To_d3-To_n)/dmax(3);

%Optimal sensitivity matrix
F = [(yd1-yn)/delta_d1 (yd3-yn)/delta_d3]; %(yd2-yn)/delta_d2 

F = F([1,3,5,end],:); % -(ts + qhr + s_qrt) -d2
%F = F([3,5,end],:); % - (ts + qhr + s_qrt + To) -d2

%F = F([1,3:5,end],:); % - (ts + t_sqrt)
%F = F([1:3,5:end],:); % -qhr
%F = F([1,3,5:end],:); % - ( qhr + tswitch )
%F = F([2:3,5:end],:); % - (qhr + To)
%F = F([3,5:end],:); % - (qhr + To + tswitch)

%Nullspace method, HF = 0
H = null(F');
H = H'
H1 = [0 0]'; %- To
H2 = [0 0]'; %- ts
H4 = [0 0]'; % - qhr
H6 = [0 0 ]'; % - t_sqrt
H = [H(:,1) H2 H(:,2) H4 H(:,3) H6 H(:,4)] % -(ts + qhr + s_qrt) -d2
% H = [H1 H2 H(:,1) H4 H(:,2) H6 H(:,3)] % -(ts + qhr + s_qrt + To) -d2
% H_rad = [0 0 0 0 0 0 0] % -(ts + qhr + s_qrt + To) -d2
% H = [H(1,:) ; H_rad] % -(ts + qhr + s_qrt + To) -d2

%H = [H(:,1) H2 H(:,2) H(:,3) H(:,4) H6 H(:,5)] % - (ts + t_sqrt) 
    
%H = [H(:,1) H(:,2) H(:,3) H4 H(:,4) H(:,5) H(:,6)] %-qhr
%H = [H(:,1) H2 H(:,2) H4 H(:,3) H(:,4) H(:,5)] % - ( qhr + tswitch )
%H = [H1 H(:,1) H(:,2) H4 H(:,3) H(:,4) H(:,5)] % -(qhr + To)
%H = [H1 H2 H(:,1) H4 H(:,2) H(:,3) H(:,4)] % -(qhr + T+ + tswitch)

%% Simulation and plotting

% H1 = H(1,:);
% c1_1 = H1*yd1; 
% c2_1 = H1*yd2;
% cn_1 = H1*yn;
% 
% H2 = H(2,:);
% c1_2 = H2*yd1; 
% c2_2 = H2*yd2;
% cn_2 = H2*yn;
% 
% H3 = H(3,:);
% c1_3 = H3*yd1;
% c2_3 = H3*yd2;
% cn_3 = H3*yn;

To = 273+0
sim('simulink_simple_PI_nullspace')%,'AlgebraicLoopSolver','LineSearch')


%Simulation time
t = time_1; %[h] 1 day

%Output. The different controlled variables
c1 = c(:,1);
c2 = c(:,2);
% c3 = c(:,3);
% c4 = c(:,4);
% c5 = c(:,5);

%Plotting of the different controlled variables

figure(10)
subplot(4,1,1)
plot(t,c1,'r')
xlabel('time [h]')
ylabel('controlled variable 1')
legend('t + DeltaT') 

subplot(4,1,2)
plot(t,c2,'r')
xlabel('time [h]')
ylabel('controlled variable 2')
legend('To + t') 

% subplot(4,1,3)
% plot(t,c3,'r')
% xlabel('time [h]')
% ylabel('controlled variable 3')
% legend('t + distance') 

% subplot(4,1,4)
% plot(t,c4,'r')
% xlabel('time [h]')
% ylabel('controlled variable 4')
% legend('t + distance + t sqrt') 

% subplot(5,1,5)
% plot(t,c4,'r')
% xlabel('time [h]')
% ylabel('controlled variable 5')
% legend('t + distance + t sqrt + tsq') 


%Matrix B (=d) from optimization 
d1_1 = [ones(n1,1)*(Tmin_a)-273; 
     ones(n2,1)*(Tmin_b)-273;
     ones(n3,1)*(Tmin_b)-273;
     ones(n4,1)*(Tmin_b)-273; 
     ones(n5,1)*(Tmin_b)-273]; % [C]

% Plot of Tr, Tr_set point and Tr_constraint 

%Outputs
Tr_1 = simout_1(:,2)-273; % [C]
Tr_set_1 = Set_out_1-273; % [C]
qhr_out_1 = Qhr;
t_1 = time_1;

% Plotting 

figure(11)
Subplot(2,1,1)
plot(t_1,Tr_1,'r',t_1,Tr_set_1,'b',t1,d1_1,'g')
ylabel('Temperature [C]')
xlabel('Time [h]')
legend ('Tr','Tr set', 'Tr constraint') 

Subplot(2,1,2)
plot(t_1,qhr_out_1)
ylabel('qhr [J/h]')
xlabel('Time [h]')
title('Room')

% Loss

% Loss, L = Js - Jopt / Jopt

Jopt = x_E(480)  %value at time 9.6, which is the last time before the temperature decreases. 
Js = E_out_1(end)

%Loss
L = ((Js-Jopt)/Jopt)*100 



